from RoboidControl.Participant import Participant
from RoboidControl.Thing import Thing
from RoboidControl.Messages.ParticipantMsg import ParticipantMsg
from RoboidControl.Messages.NetworkIdMsg import NetworkIdMsg
from RoboidControl.Messages.ThingMsg import ThingMsg
from RoboidControl.Messages.NameMsg import NameMsg
from RoboidControl.Messages.ModelUrlMsg import ModelUrlMsg
from RoboidControl.Messages.Messages import IMessage
from RoboidControl.Messages import *

import socket
import threading
import time
from typing import Optional

# import sys
# micropython = 'micropython' in sys.modules

# if micropython:
#     from MicroPython.uPythonParticipant import uPythonParticipant

class ParticipantUDP(Participant):
    """! A local participant is the local device which can communicate with other participants.
    
    It manages all local things and communcation with other participants. Each application has a local participant which is usually explicit in the code.
    An participant can be isolated. In that case it is standalong and does not communicate with other participants. 

    Currently, only UDP communication is supported
    """

    #buffer = None
    nextPublishMe = 0
    #others = None
    #thread = None
    name = "Participant"
    isolated_participant = None

    def __init__(self, 
                 port: int = 7681,
                 ip_address: Optional[str] = None,
                 local_port: int = 7681
                 ) -> None:
        super().__init__(ip_address = "127.0.0.1", port = local_port)

        # if local_port == 0:
        #     local_port = port
        # self.local_port = local_port

        ## True if the participant is running isolated.
        # Isolated participants do not communicate with other participants
        self.is_isolated: bool = True
        self.remote_site: Optional[Participant] = None

        ## The other participants communicating with this participant
        self.others: list[Participant] = []

        if port != 0:
            self.is_isolated = False
            if ip_address is not None:
                self.remote_site = Participant(ip_address, port)
                self.others.append(self.remote_site)

        self.buffer: bytearray = bytearray(256)
        self.publishInterval = 3000 # 3 seconds

        self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)        
        self.udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1) 
        self.udp_socket.bind(("0.0.0.0", self.port))

        self.thread = threading.Thread(target = self.Receiver)
        self.thread.daemon = True
        self.thread.start()
    
    @staticmethod
    def Isolated() -> 'ParticipantUDP':
        if ParticipantUDP.isolated_participant == None:
            ParticipantUDP.isolated_participant = ParticipantUDP(0)
        return ParticipantUDP.isolated_participant

#region Update

    def GetParticipant(self, ip_address: str, port: int):
        # print(f'{self.name} Get participant {ip_address} {port}')
        # for item in self.others:
        #     print(f'- {item.ip_address} {item.port}')

        found_participants = (item for item in self.others
                               if item.ip_address == ip_address and item.port == port)
        participant = next(found_participants, None)
        return participant

    def AddParticipant(self, ip_address: str, port: int):
        print(f'{self.name} Add participant {ip_address} {port}')
        remote_participant = Participant(ip_address = ip_address, port = port)
        self.others.append(remote_participant)
        return remote_participant

    def Update(self, currentTimeMs: Optional[int] = None):
        if currentTimeMs is None:
            currentTimeMs = int(time.time() * 1000)

        if self.is_isolated == False:
            if self.publishInterval > 0 and currentTimeMs > self.nextPublishMe:
                msg = ParticipantMsg(self.network_id)
                if self.remote_site == None:
                    self.Publish(msg)
                else:
                    self.Send(self.remote_site, msg)
                    
                self.nextPublishMe = currentTimeMs + self.publishInterval

        self.UpdateMyThings(currentTimeMs)
    
    def UpdateMyThings(self, currentTimeMs: int):
        for thing in self.things:
            # if thing.hierarchyChanged and not (self.isIsolated or self.network_id == ):
            #     thingMsg = ThingMsg(self.network_id, thing)
            #     self.Send(self.remote_site, thingMsg)
            if thing.owner is not None:
                poseMsg = PoseMsg(thing.owner.network_id, thing)
                if self.remote_site is not None:
                    self.Send(self.remote_site, poseMsg)

            thing.Update(currentTimeMs, False)
            if not(self.is_isolated or self.network_id == 0):
                # if thing.terminate:
                #     destroyMsg = DestroyMsg(self.network_id, thing)
                #     self.Send(self.remote_site, destroyMsg)
                # else:
                    if self.remote_site is not None and thing.owner is not None:
                        # Send to remote site
                        poseMsg = PoseMsg(thing.owner.network_id, thing)
                        self.Send(self.remote_site, poseMsg)
                        binaryMsg = BinaryMsg(thing.owner.network_id, thing)
                        self.Send(self.remote_site, binaryMsg)
            # if thing.terminate:
            #     self.Remove(thing)

#endregion

#region Send

    def SendThingInfo(self, owner: Participant, thing: 'Thing', recursively: bool = False):
        self.Send(owner, ThingMsg(self.network_id, thing))
        self.Send(owner, NameMsg(self.network_id, thing))
        self.Send(owner, ModelUrlMsg(self.network_id, thing))
        self.Send(owner, PoseMsg(self.network_id, thing, True))
        self.Send(owner, BinaryMsg(self.network_id, thing))

        if recursively:
            for child in thing.children:
                self.SendThingInfo(owner, child, recursively)


    def Send(self, owner: Participant, msg: IMessage):
        buffer_size = msg.Serialize([self.buffer])
        if buffer_size <= 0:
            return True
        
        # print(f'{self.name} send {self.buffer[0]} to {owner.ip_address} {owner.port}')
        self.udp_socket.sendto(self.buffer[:buffer_size], (owner.ip_address, owner.port))
        return True

    def Publish(self, msg: IMessage):
        buffer_size = msg.Serialize([self.buffer])
        if buffer_size <= 0:
            return True
        
        # print(f'publish {self.buffer[0]} to {self.port}')
        self.udp_socket.sendto(self.buffer[:buffer_size], ('<broadcast>', self.port))
        return True

#endregion       

#region Receive
    
    def Receiver(self):
        while True:
            try:
                data, addr = self.udp_socket.recvfrom(1024)
                remote_ip_address = addr[0]
                remote_port = addr[1]
                # print(f'msg received from {remote_ip_address}:{remote_port}')
                remote_participant = self.GetParticipant(remote_ip_address, remote_port)
                if remote_participant is None:
                    # print(f'new participant')
                    remote_participant = self.AddParticipant(remote_ip_address, remote_port)
                self.ReceiveData(data, remote_participant)
            except ConnectionError:
                self.network_id = 0
                pass

    def ReceiveData(self, data: bytes, sender: Participant):
        msgId = data[0]
        # print(f'msg {msgId} ')
        match msgId:
            case ParticipantMsg.id:
                self.ProcessParticipantMsg(sender, ParticipantMsg(data))
            case NetworkIdMsg.id:
                self.ProcessSiteIdMsg(sender, NetworkIdMsg(data))
            # case InvestigateMsg.id:
            #     self.ProcessInvestigateMsg(InvestigateMsg(data))
            case ThingMsg.id:
                self.ProcessThingMsg(ThingMsg(data))
            case NameMsg.id:
                self.ProcessNameMsg(NameMsg(data))
            case ModelUrlMsg.id:
                self.ProcessModelUrlMsg(ModelUrlMsg(data))
            case BinaryMsg.id:
                self.ProcessBinaryMsg(BinaryMsg(data))            
            case _:
                pass

    def ProcessParticipantMsg(self, sender: Participant, msg: ParticipantMsg):
        pass

    def ProcessSiteIdMsg(self, sender: Participant, msg: NetworkIdMsg):
        print(f'{self.name} Process SiteMsg {self.network_id} -> {msg.network_id}')
        if self.network_id != msg.network_id:
            self.network_id = msg.network_id
            for thing in self.things:
                #if thing.parent is None:
                self.SendThingInfo(sender, thing, recursively=True)

    def ProcessInvestigateMsg(self, data: bytearray):
        pass

    def ProcessThingMsg(self, msg: ThingMsg):
        print(f'received thing {msg.thing_id}')

    def ProcessNameMsg(self, msg: NameMsg):
        print(f'received name {msg.name}')
    
    def ProcessModelUrlMsg(self, msg: ModelUrlMsg):
        print(f'received model url: {msg.url}')

    def ProcessBinaryMsg(self, msg: BinaryMsg):
        # print('received binary data')
        thing: Optional[Thing] = self.Get(msg.thing_id)
        if thing is not None:
            thing.ProcessBinary(msg.data)

#endregion